/* ************************************************************************** *
 *  Filename: Agent.h
 *
 *  This client program is developed based on a sample client.
 *  Client Program (Header file).
 *  by Neo Say Poh/ Eldwin Tan Say Lin. 
 * ************************************************************************** */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>

#define FLAG_COUNT			   53
#define GOAL_COUNT			   	2
#define	LINE_COUNT				4
#define FIELD_LENGTH		 52.5
#define FIELD_WIDTH		 	 34.0
#define PENALTY_AREA_LENGTH	 16.5
#define PENALTY_AREA_WIDTH  20.16
#define GOAL_WIDTH			14.02
#define CENTER		 	      0.0
#define BUFSIZE 	 	  	  256
#define Bool		 	  	  int
#define TRUE 	  	       	    1
#define FALSE	  	   	    	0
#define INIT_OK			   	"init"
#define BEFORE_KICK_OFF	 	"before_kick_off"

/* Kickable Range */
#define BALL_SIZE			0.085
#define PLAYER_SIZE			0.8
#define KICKABLE_RANGE		(1.0+BALL_SIZE+PLAYER_SIZE)

/* Sensor Information */
#define SEE 				"see"
#define HEAR				"hear"
#define SENSE_BODY			"sense_body"
#define ERROR				"error"
/*** Hear direction ***/
#define SELF				"self"
#define REFEREE				"referee"
#define PLAYER				"player"

#define DEFAULT_PORT_NUMBER	 6000
#define MAXMESG			 	 1024

#ifdef MEMSET
#define bzero(a, b) memset(a, 0, b)
#endif

typedef struct socket
{
	int	socketfd  ;
    	struct sockaddr_in	serv_addr ;
} Socket ;

#define  MAX_TIME_COUNTER 	 	1000
#define  PI 					(3.1415927)
#define  Get_dist(x1,y1,x2,y2)  (sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)))
#define  Random_num(num) 		(random()%num)
#define  Abs_d(d) 				((d>=0.0) ? (d) : (-d))

struct relative_address
{
	int  count; /* time counter */
	double dist;
	int  dir,vdir;
	double d[4]; /* 0==>dist,1==>dir, 2==>dist_chg,3==>dir_chg */
};

struct player_relative_address
{
	int  count; /* time counter */
	double dist;
	int  dir,vdir;
	double d[5]; /* 0==>dist,1==>dir, 2==>dist_chg,3==>dir_chg,4==>face_dir */
};

struct sequence_list
{
	int  count;
	double dist[4],dir[4];
};

/* All visible flags */
#define FLAG_STRINGS {"flag c", "flag c t",\
                      "flag l 0", "flag l t", "flag l b",\
                      "flag p l c", "flag p l t", "flag p l b",\
                      "flag g l t", "flag g l b",\
                      "flag l t 10", "flag l b 10", "flag l t 20",\
                      "flag l b 20", "flag l t 30", "flag l b 30",\
                      "flag t l 10", "flag t l 20", "flag t l 30",\
                      "flag t l 40", "flag t l 50",\
                      "flag b l 10", "flag b l 20", "flag b l 30",\
                      "flag b l 40", "flag b l 50",\
                      "flag t 0", "flag c b",\
                      "flag r 0", "flag r b", "flag r t",\
                      "flag p r c", "flag p r b", "flag p r t",\
                      "flag g r b", "flag g r t",\
                      "flag r b 10", "flag r t 10", "flag r b 20",\
                      "flag r t 20", "flag r b 30", "flag r t 30",\
                      "flag b r 10", "flag b r 20", "flag b r 30",\
                      "flag b r 40", "flag b r 50",\
                      "flag t r 10", "flag t r 20", "flag t r 30",\
                      "flag t r 40", "flag t r 50",\
		      "flag b 0", NULL}

typedef enum
{
	flagC, flagCT,
	flagL0, flagLT,	flagLB,
	flagPLC, flagPLT, flagPLB,
	flagGLT, flagGLB,
	flagLT10, flagLB10, flagLT20, flagLB20, flagLT30, flagLB30,
	flagTL10, flagTL20, flagTL30, flagTL40, flagTL50,
	flagBL10, flagBL20, flagBL30, flagBL40, flagBL50,
	flagT0, flagCB,
	flagR0, flagRB, flagRT,
	flagPRC, flagPRB, flagPRT,
	flagGRB, flagGRT,
	flgRB10, flagRT10, flagRB20, flagRT20, flagRB30, flagRT30,
	flagBR10, flagBR20, flagBR30, flagBR40, flagBR50,
	flagTR10, flagTR20, flagTR30, flagTR40, flagTR50, flagB0,
} FlagIndex;

/* ********************* Functions in netif.c ******************************* */
extern Socket init_connection(char* host, int port) ;
extern Bool   send_message(Socket sock, char* buf) ;
extern int    receive_message(Socket *sock, char *buf, int size) ;
extern void   close_connection(Socket sock) ;
/* ************************************************************************** */

/* ******************** Functions in control_com.c ************************** */
extern Bool control_com_kick(Socket *sock, int pow, double dir);
extern Bool control_com_dash(Socket *sock, int pow);
extern Bool control_com_turn(Socket *sock, double moment);
/* ************************************************************************** */

/* ******************** Functions in misc.c ********************************* */
extern double get_angle(double dx, double dy);
extern Bool   get_my_posi(void); 
extern char   *get_para(struct relative_address *st,char *buf);
extern char   *get_player_para(struct player_relative_address *st,char *buf);
extern Bool   init_condition(Socket *sock, char *t_name, int goalie_flag);
/* ************************************************************************** */
