/* ************************************************************************** *
 *  Filename: 	Agent.c
 *  Functions: 	get_theta(), pass(), search_ball(), go(), go_abs(),
 *	        go_predict(), gk_think(), think(), see(), hear(),
 *	   	message_loop().
 *
 *  This client program is developed based on a sample client.
 *  Client Program (Main program).
 *  by Neo Say Poh/Eldwin Tan Say Lin.
 * ************************************************************************** */

#include "Agent.h"
#define  DETECT_ANGLE  (20)	/* used in pass(). */
#define  DASH_RATIO    (1.7)

/* ****************** Functions Prototypes ********************************** */
double get_theta(double ox, double oy);
void   pass(Socket *sock, double *address, double *address2);
void   search_ball(Socket *sock);
void   go(Socket *sock, double dist, double odir);
void   go_abs(Socket *sock, double ox, double oy);
void   go_predict(Socket *sock);
int    gk_think(Socket *sock, char *t_name);
int    think(Socket *sock, char *t_name);
void   see(char *buf, char *t_name);
Bool   hear(char *buf, char *t_name);
void   message_loop(Socket sock, char *t_name, FILE* fpin, FILE* fpout);
/* ************************************************************************** */

static int 	time_stamp=0;
int    		s_flag=0,h_flag=0,b_flag=0,p_flag=0,collision_flag=0,move_flag=0;
int    		list_flag=0;
int    		catch_flag=0; 		/* ball catched or not */

extern double 	mx,my,mdir;
extern double 	plocation[2];
extern double 	bx,by,blocation[2];
extern int    	player;
extern char   	side;
char  	  	cir_flag=0;	/* 0-Search ball, 1-go to off_posi */
extern int    	time_counter;
int    	  	mp_num,op_num; /* mp_num - my player's num, op_num - opponent player's num */

/* Positions */
extern int 	init_posi[11][10];
extern double 	flag_posi[53][2],goal_posi[2][2];
double 		off_posi[11][2]={{-51,0},{-30,0},{-25,-17},{-25,17},{-10,0},{5,-10},
                        	{5,10},{16,-23},{16,23},{23,-10},{23,10}};
double 		area_posi[11][2]={{14,20},{15,15},{28,17},{28,17},{20,10},{20,10},
                         	{20,10},{38,12},{38,12},{23,15},{23,15}}; 

struct relative_address ball;
struct player_relative_address mplayer[11],oplayer[11];
extern struct relative_address flag[FLAG_COUNT],goal[GOAL_COUNT],line[LINE_COUNT];
extern struct sequence_list s_list; /* used for getting out of collision... */

/* ************************************************************************** *
 * get_theta() - get relative angle.
 *
 * Args:
 *		ox : x-coordinates of destination.
 *		oy : y-coordinates of destination.
 * Return type:
 *		double : angle difference between current pos. & destination.
 * ************************************************************************** */
double get_theta(double ox,double oy)
{
    	double  odir;
	
	odir=get_angle(ox-mx,oy-my);
    	odir-=mdir;
    	if (odir<-180)
		odir+=360;
    	else if (odir>180)
		odir-=360;
    	return odir;
}

/* ************************************************************************** *
 * pass() - pass to teammates or shoot at goal.
 * 
 * Args:
 *		sock     : Socket.
 *		address  : Address of the destination coordinates.
 *		address2 : Address of alternative dest. coordinates.
 * Return type:
 * 		void.
 * ************************************************************************** */
void pass(Socket *sock, double *address, double *address2)
{
    	double 	dist,ox=address[0],oy=address[1];
    	double 	dir2,dir,odir;
    	int  	i,pow=0,k_flag=0;

    	if (address2==NULL)
		address2=address;
    	dist=Get_dist(ox,oy,mx,my);
    	dir=get_theta(ox,oy);
    	if ((my>33.0) || (my<-33.0)) /* around side line */
	{
		if (my>33.0)
		{
		    	if ((mdir>0) && (mdir<90))
				k_flag=-1;
		    	else if ((mdir>90) && (mdir<180))
				k_flag=1;
		} 
		else
		{
		    	if ((mdir>180) && (mdir<270))
				k_flag=-1;
		    	else if ((mdir>270) && (mdir<360))
				k_flag=1;
		}
    	}
    	if (dir>120) /* kick back */
		k_flag=1;
    	else if (dir<-120)
		k_flag=-1;

    	if (k_flag!=0)
	{
		control_com_kick(sock,8,105*k_flag);
		usleep(20000);
		control_com_turn(sock,90*k_flag);	    
		dir-=70*k_flag;
		usleep(210000);
    	}

    	if (((address[0]==goal_posi[1][0]) && (address[1]==goal_posi[1][1])) || (player==0)) 
		dist+=30.0; 	/* increase pass/shoot power */
    	else if (k_flag==0) /* enemy detection */
	{
		dir2=get_theta(address2[0],address2[1]);
		
		if (dir==dir2)
		    	k_flag=1;
		else if ((dir2<120) && (dir2>-120)) /* check the other pass */
		{
		    	for (i=0;i<op_num;i++)
			{
				if (oplayer[i].dist<20.0)
				{
		    			odir=oplayer[i].dir-dir2;
		    			if ((odir>-DETECT_ANGLE) && (odir<DETECT_ANGLE))
						k_flag=1;
				}	
	    		}
		} 
		else
		    	k_flag=1;

		for (i=0;i<op_num;i++) /* check original pass */
		{
	    		if (oplayer[i].dist<20.0)
			{
				odir=oplayer[i].dir-dir;
				if ((odir>-DETECT_ANGLE) && (odir<DETECT_ANGLE))
				{
		    			if (!k_flag)
						dir=dir2;
					else
					{
						if (odir>0)
				    			odir=DETECT_ANGLE-odir;
						else
				    			odir=-DETECT_ANGLE-odir;
						dir-=odir;
		    			}
				}
	    		}
		}
    	}
	control_com_kick(sock,(int)(dist*1.6)+20,dir);
}

/* ************************************************************************** *
 * search_ball() - search for ball by turning to ball's direction.
 * 
 * Arg:
 *		sock : Socket.
 * Return type:
 *		void.
 * ************************************************************************** */
void search_ball(Socket *sock)
{
    	int  	dir;

    	dir=ball.d[1];
    	control_com_turn(sock,dir);
}

/* ************************************************************************** *
 * go() - go to relative position indicated by dist & odir.
 *
 * Args:
 *		sock : Socket.
 *		dist : distance.
 *	 	odir : direction.
 * Return type:
 *		void.
 * ************************************************************************** */
void go(Socket *sock,double dist,double odir)
{
    	int  		pow,i;
    	static double 	tx=0.0,ty=0.0;
    	static int 	f_count=0;

    	if (Get_dist(mx,my,tx,ty)<0.5) /* collision check */
	{
		if (++f_count>6) /* stayed at same place for 6 consecutive cycles */
		{
	    		i=Random_num(2)*2-1;

	    		control_com_turn(sock,90.0*i); /* turn 90 deg. left or right randomly */
	    		list_flag=1;
	    		s_list.count=1;
	    		s_list.dist[1]=5.0; s_list.dir[1]=0.0; 		/* dash out of collision */
	    		s_list.dist[0]=0.0; s_list.dir[0]=-90.0*i;	/* turn back to original direction */
	    		f_count=0;
	    		return;
		}
    	} 
	else
		f_count=0;

	if (Abs_d(odir)>3.0) /* turn */
		control_com_turn(sock,odir);
	else /* dash */
	{
		if (dist>DASH_RATIO)
		{
	    		control_com_dash(sock,100);
	    		usleep(60000);
	    		dist-=DASH_RATIO;
		}
		pow=(int)(dist*100.0/DASH_RATIO);

		control_com_dash(sock,pow);
    	}
    	tx=mx; ty=my;
}

/* ************************************************************************** *
 * go_abs() - go back to my absolute position (ox,oy).
 *
 * Args:
 * 		sock : Socket.
 *		ox   : X-coordinate of destination.
 *		oy   : Y-coordinate of destination.
 * Return type:
 *		void.
 * ************************************************************************** */
void go_abs(Socket *sock,double ox,double oy)
{
    	double 	dist,odir; /* dist. & dir from (mx,my) to (ox,oy). */
	
   	dist=Get_dist(mx,my,ox,oy);
    	odir=get_theta(ox,oy);
	
    	go(sock,dist,odir);
}


/* ************************************************************************** *
 * go_predict() - go to ball using prediction. 
 *
 * Arg:
 *		sock : Socket.
 * Return type:
 * 		void.
 * ************************************************************************** */
void go_predict(Socket *sock)
{
    	double 	dist=ball.d[0],dir=ball.d[1];
    	double 	ddist=ball.d[2],ddir=ball.d[3]*PI/180.0;
    	double 	d,gamma;
    	double 	t;
    	int  	i;

    	if (ddir!=0.0)
	{
		t=cos(ddir)-dist/(dist-ddist);
		gamma=atan(t/sin(ddir)); 
	/*	gamma=atan((cos(ddir)-dist/(dist-ddist))/sin(ddir)); */

		if ((gamma*ddir>0.0) && (ddist>0.0)) /* the ball is approaching to me */
		{
	    		d=dist*cos(gamma);
	    		list_flag=1;
	    		s_list.count=1;
	    		i=(int)(d/DASH_RATIO)+1;
	    		if (i>3) 
				i=3;
	    		while (i>1)
			{
				s_list.dist[i]=DASH_RATIO; s_list.dir[i]=0.0;
				i--;
				d-=DASH_RATIO;
	    		}
	    		s_list.dist[1]=d; s_list.dir[1]=0.0;
	    		s_list.dist[0]=0.0; s_list.dir[0]=90.0*((ddir>0.0) ? 1.0 : -1.0);
	    		gamma=gamma*180.0/PI;
		} 
		else
		{
	    		gamma=0.0;
	    		d=dist+ddist;
		}
    	} 
	else
	{	
		gamma=0.0;
		d=dist+ddist;
    	}

    	go(sock,dist,dir+gamma);
}

/* ************************************************************************** *
 * gk_think() - determine goal-keeper`s action(s) to take.
 *
 * Args:
 *		sock   : Socket.
 *		t_name : team name.
 * Return type:
 *		int : 1 - Successful.
 *	  		  0 - No actions taken.
 * ************************************************************************** */
int  gk_think(Socket *sock, char *t_name)
{
    	static int 	rotate_flag=0;
	int  		i;
	char 		str[50];
    	double 		dx,dy,dist,odir;
	char 		mesg[30]; 			/* move message & server's response */
	char 		buf[MAXMESG];
	int  		rand_num1, rand_num2;		/* random numbers for passing purpose */
	double 		dir;
	int  		c_dir;				/* direction for catching ball */
	double 		catch_x_dist, catch_y_dist;

    	if (!s_flag) /* see flag 1... get info this time,0... not yet */
	{
		if (move_flag) /* half time or time extended */
		{
			do
		 	{
				sprintf(mesg, "(move %d %d)\n", init_posi[player][0], init_posi[player][1]);
				usleep(100000);
				send_message(*sock, mesg);

				while (receive_message(sock, buf, MAXMESG)<=0);
            			sscanf(buf, "(%s ", mesg);

				if (!strcmp(mesg, SEE))
                			see(buf,t_name);
				else if (hear(buf, t_name)) /* hear referee's msg: kick_off_? */
                   		 	break;
            			else
                		continue;

        		} while ((!(get_my_posi())) || (mx>(init_posi[player][0]+3))
				        || (mx<(init_posi[player][0]-3))
			        	|| (my>(init_posi[player][1]+3)) || (my<(init_posi[player][1]-3))); /* move to init_posi until successful */

			while(1) /* wait for referee to restart match */
        		{
            		while (receive_message(sock, buf, MAXMESG)<=0);

            		sscanf(buf, "(%s ", mesg);
            		if (!strcmp(mesg, HEAR))
            			{
                		if (hear(buf, t_name)) /* hear referee's msg: kick_off_? */
                    			break;
            			}
        		}
			move_flag=0;
		}
		return FALSE;
	}

   	if (!(get_my_posi())) /* get my absolute position by using line(s) */
  	{
		control_com_turn(sock,180); /* when lost */
		list_flag=1;
		s_list.count=0;
		s_list.dist[0]=DASH_RATIO; s_list.dir[0]=0.0;
		return TRUE;	
    	}
	
	if (list_flag) /* sequential list exists */
	{
		i=s_list.count;
		
		go(sock,s_list.dist[i],s_list.dir[i]);
		if (i==0)
	    	list_flag=0;
		else
	    	s_list.count--;
		return TRUE;
    	}

    	if (b_flag) /* see ball info. */
	{
		odir=mdir+ball.d[1];
		bx=mx+ball.d[0]*cos(odir/180.0*PI);
		by=my+ball.d[0]*sin(odir/180.0*PI);
		sprintf(str,"(say %s %d %3.1lf %3.1lf)\n",t_name,time_stamp+1,bx,by);
		rotate_flag=0; /* rotate_flag describes searching state */
   	} 
	else if (h_flag) /* hear ball info. */
	{
		ball.count=time_counter;
		ball.d[0]=Get_dist(mx,my,bx,by);
		ball.d[1]=get_theta(bx,by);
		ball.d[2]=0.0;
		ball.d[3]=0.0;
		ball.dist=ball.d[0];
		rotate_flag=0;
		sprintf(str,"(say %s %d %3.1lf %3.1lf)\n",t_name,time_stamp+1,bx,by);
    	}

	/* determine the ball distance. */
	catch_x_dist=bx-mx;
	catch_y_dist=by-my;

    	if (cir_flag) /* cir_flag=1... go to my position,0... search ball */
	{
		if ((ball.count==time_counter) && ((ball.d[0]<10.0)
			|| ((bx<(off_posi[player][0]+area_posi[player][0]))
			&& (bx>(off_posi[player][0]-area_posi[player][0]))
			&& (by<(off_posi[player][1]+area_posi[player][1]))
			&& (by>(off_posi[player][1]-area_posi[player][1]))))) /* ball in my area_posi */
		{
//			if (ball.d[0]!=mplayer[mp_num].d[0] && ball.d[1]!=mplayer[mp_num].d[1])
//			{
		    		go_predict(sock);
				cir_flag=0;
//			}
//			else
				go_abs(sock,off_posi[player][0],off_posi[player][1]);
		}
		else if (Get_dist(off_posi[player][0],off_posi[player][1],mx,my)>6.0)
		{
			send_message(*sock, str);
//	    		go_abs(sock,off_posi[player][0],off_posi[player][1]);
			if(by>=15.0)
				go_abs(sock,-48,7);
			else if(by<=(-15.0))
				go_abs(sock,-48,-7);
			else if(by>(-15.0) && by<15.0)
				go_abs(sock,-48,0);
		}
		else
		{
		    	cir_flag=0;
		    	if ((b_flag) && (Abs_d(ball.d[1])<1.0))
				send_message(*sock, str);
		    	else
				search_ball(sock);
		}
	}
	else if (ball.count!=time_counter) /* can't see the ball, so search it */
	{
//		go_abs(sock,off_posi[player][0],off_posi[player][1]);
		if (rotate_flag==0)
		{
		    	control_com_turn(sock,-45.0);
		    	rotate_flag=1;
		} 
		else if (rotate_flag==1)
		{
			control_com_turn(sock,90.0);
		    	rotate_flag=2;
		} 
		else
		    	control_com_turn(sock,45.0); 
		search_ball(sock);
    	} 
	else if ((bx<off_posi[player][0]-area_posi[player][0])
		     || (bx>off_posi[player][0]+area_posi[player][0])
		   	 || (by<off_posi[player][1]-area_posi[player][1])
		     || (by>off_posi[player][1]+area_posi[player][1])) /* out of player's movement range */
	{
		if (ball.d[0]<=45.0)
		{
			if(by>=15.0)
				go_abs(sock,-48,7);
			else if(by<=(-15.0))
				go_abs(sock,-48,-7);
			else if(by>(-15.0) && by<15.0)
				go_abs(sock,-48,0);
		}
		else if (Get_dist(off_posi[player][0],off_posi[player][1],mx,my)>2.0)
		{
			send_message(*sock, str);
	    		go_abs(sock,off_posi[player][0],off_posi[player][1]); /* go back to my off_posi */
/*			if(by>=15.0)
				go_abs(sock,-48,7);
			else if(by<=(-15.0))
				go_abs(sock,-48,-7);
			else if(by>(-15.0) && by<15.0)
				go_abs(sock,-48,0); */
	    		cir_flag=1;
		} 
		else
		{
			if ((b_flag) && (Abs_d(ball.d[1])<1.0))
			{
//	    			go_abs(sock,off_posi[player][0],off_posi[player][1]); /* go back to my off_posi */
/*				if(by>=15.0)
					go_abs(sock,-48,7);
				else if(by<=(-15.0))
					go_abs(sock,-48,-7);
				else if(by>(-15.0) && by<15.0)
					go_abs(sock,-48,0); */
				send_message(*sock, str);
			}
	    		else
				search_ball(sock);		
		}
	}
	else 
	{
		if (!catch_flag) /* catch ball */
		{
			catch_flag=TRUE;
			if ((Abs_d(catch_x_dist)>2) || (Abs_d(catch_y_dist)>1))
				go_predict(sock);
			else
			{
				dir=get_theta(bx,by);
				c_dir=(int)dir;
				sprintf(str, "(catch %d)\n",c_dir);
				usleep(100000);
       			     	send_message(*sock, str);
			}
		}
		else /* pass ball to teammate */
		{
//			if (Get_dist(mx, my, plocation[0], plocation[1])<oplayer[op_num].d[0])
//				pass(sock,plocation,NULL);
//			else
//			{
				/* generating random numbers for passing */
				rand_num1=Random_num(6);
//				rand_num2=Random_num(4);
//				if (rand_num2>=rand_num1)
//					rand_num2++;
				pass(sock,off_posi[rand_num1+1],NULL);
//        	  		pass(sock,off_posi[rand_num1+4],off_posi[rand_num2+4]); /* pass to mf */
//			}
			go_abs(sock,off_posi[player][0],off_posi[player][1]); /* go back to my off_posi */
			catch_flag=FALSE;
		}
    	}
    	return TRUE;
}	

/* ************************************************************************** *
 * think() - determine the action(s) to take by players.
 *
 * Args:
 *		sock   : Socket.
 *		t_name : team name.
 * Return type:
 *		int : 1 - Successful.
			  0 - No actions taken.
 * ************************************************************************** */
int  think(Socket *sock, char *t_name)
{
    	static int 	rotate_flag=0;
	int  		i;
	char 		str[50];
    	double 		dx,dy,dist,odir;
	char 		mesg[30]; 		/* move message & server's response */
	char 		buf[MAXMESG];
	int  		rand_num1, rand_num2;	/* random numbers for passing purpose */

    	if (!s_flag) /* see flag 1... get info this time,0... not yet */
	{
		if (move_flag) /* half time or time extended */
		{
			do
 			{
				sprintf(mesg, "(move %d %d)\n", init_posi[player][0], init_posi[player][1]);
				usleep(100000);
				send_message(*sock, mesg);

				while (receive_message(sock, buf, MAXMESG)<=0);
       			     	sscanf(buf, "(%s ", mesg);

				if (!strcmp(mesg, SEE))
                			see(buf,t_name);
				else if (hear(buf, t_name)) /* hear referee's msg: kick_off_? */
                   			 break;
            			else
                			continue;

        		} while ((!(get_my_posi())) || (mx>(init_posi[player][0]+3))
					 || (mx<(init_posi[player][0]-3))
					 || (my>(init_posi[player][1]+3)) || (my<(init_posi[player][1]-3))); /* move to init_posi until successful */

			while(1) /* wait for referee to restart match */
        		{
            			while (receive_message(sock, buf, MAXMESG)<=0);

            			sscanf(buf, "(%s ", mesg);
            			if (!strcmp(mesg, HEAR))
            			{
                			if (hear(buf, t_name)) /* hear referee's msg: kick_off_? */
                    				break;
            			}
        		}
			move_flag=0;
		}
		return FALSE;
	}

   	if (!(get_my_posi())) /* get my absolute position by using line(s) */
  	{
		control_com_turn(sock,180); /* when lost */
		list_flag=1;
		s_list.count=0;
		s_list.dist[0]=DASH_RATIO; s_list.dir[0]=0.0;
		return TRUE;	
    	}
	
	if (list_flag) /* sequential list exists */
	{
		i=s_list.count;
		
		go(sock,s_list.dist[i],s_list.dir[i]);
		if (i==0)
	    	list_flag=0;
		else
	    	s_list.count--;
		return TRUE;
    	}

    	if (b_flag) /* see ball info. */
	{
		odir=mdir+ball.d[1];
		bx=mx+ball.d[0]*cos(odir/180.0*PI);
		by=my+ball.d[0]*sin(odir/180.0*PI);
		sprintf(str,"(say %s %d %3.1lf %3.1lf)\n",t_name,time_stamp+1,bx,by);
		rotate_flag=0; /* rotate_flag describes searching state */
    	} 
    	else if (h_flag) /* hear ball info. */
	{
		ball.count=time_counter;
		ball.d[0]=Get_dist(mx,my,bx,by);
		ball.d[1]=get_theta(bx,by);
		ball.d[2]=0.0;
		ball.d[3]=0.0;
		ball.dist=ball.d[0];
		rotate_flag=0;
		sprintf(str,"(say %s %d %3.1lf %3.1lf)\n",t_name,time_stamp+1,bx,by);
    	}

    	if (cir_flag) /* cir_flag=1... go to my position,0... search ball */
	{
		if ((ball.count==time_counter) && ((ball.d[0]<15.0) 
			|| ((bx<(off_posi[player][0]+area_posi[player][0]))
			&& ((bx>(off_posi[player][0]-10.0)) || ((bx>(off_posi[player][0]-area_posi[player][0]))
			&& ((player==2) || (player==3) || (player==7) || (player==8))))
			&& (by<(off_posi[player][1]+area_posi[player][1]))
			&& (by>(off_posi[player][1]-area_posi[player][1]))))) /* ball in my area_posi */
		{
//			if (ball.d[0]!=mplayer[mp_num].d[0] && ball.d[1]!=mplayer[mp_num].d[1]) 
//			{
				if (bx<=30.0)
				{	
		  		  	go_predict(sock);
					cir_flag=0;
				}
//			}
//			else  
//			{
//				if (ball.d[0]==oplayer[op_num].d[0] && ball.d[1]==oplayer[op_num].d[1])  
//					go_abs(sock,bx,by);
//					go(sock,ball.d[0],ball.d[1]);
				else 
					go_abs(sock,off_posi[player][0],off_posi[player][1]);
//			}	
		}
		else
		{
	    		cir_flag=0;
	    		if ((b_flag) && (Abs_d(ball.d[1])<1.0))
				send_message(*sock, str);
	    		else
				search_ball(sock);
		}
	}
	else if (ball.count!=time_counter) /* can't see the ball, so search it */
	{
	 	go_abs(sock,off_posi[player][0],off_posi[player][1]);
		if (rotate_flag==0)
		{
	    		control_com_turn(sock,-45.0);
	    		rotate_flag=1;
		} 
		else if (rotate_flag==1)
		{
			control_com_turn(sock,90.0);
	    		rotate_flag=2;
		} 
		else
	    		control_com_turn(sock,45.0);
    	} 
	else if ((ball.dist>area_posi[player][0]) 
	        || ((bx<(off_posi[player][0]-10.0)) && ((player!=2)
		&& (player!=3) && (player!=7) && (player!=8)))
	 	|| (bx<off_posi[player][0]-area_posi[player][0])
	     	|| (bx>off_posi[player][0]+area_posi[player][0])
		|| (by<off_posi[player][1]-area_posi[player][1])
		|| (by>off_posi[player][1]+area_posi[player][1])) /* out of player's position range */
	{
		if (Get_dist(off_posi[player][0],off_posi[player][1],mx,my)>3.0)
		{
	    		send_message(*sock, str);
			go_abs(sock,off_posi[player][0],off_posi[player][1]); /* go back to my off_posi */
	    		cir_flag=1;
		} 
		else
		{
			if ((b_flag) && (Abs_d(ball.d[1])<1.0))
				send_message(*sock, str);
	    		else
				search_ball(sock);		
		}
	}
	else if (ball.dist>KICKABLE_RANGE)
		go_predict(sock);
	else
	{
		if (player==1 || player==2 || player==3)
		{
			switch(Random_num(3))
			{
				case 0:
					if (Random_num(2)==0)
						pass(sock,off_posi[4],off_posi[5]);
					else
						pass(sock,off_posi[4],off_posi[6]);
					break;
				case 1:
					pass(sock,off_posi[5],off_posi[4]);
					break;
				case 2:
					pass(sock,off_posi[6],off_posi[4]);
					break;
			}
			go_abs(sock,off_posi[player][0],off_posi[player][1]);
		}
//		else if ((my>(PENALTY_AREA_WIDTH+5)) || (my<(-PENALTY_AREA_WIDTH-5))) /* right & left flanks */
//		{
//	    		if (my>(PENALTY_AREA_WIDTH+5)) /* in the bottom area */
//				i=0;
//	    		else /* in the top area */
//				i=1;
		if (player==4)
		{
			switch(Random_num(4))
			{
				case 0:
					pass(sock,off_posi[5],off_posi[6]);
					break;
				case 1:
					pass(sock,off_posi[6],off_posi[5]);
					break;
				case 2:
					pass(sock,off_posi[7],off_posi[5]);
					break;
				case 3:
					pass(sock,off_posi[8],off_posi[6]);
					break;
			}
			go_abs(sock,off_posi[player][0],off_posi[player][1]);
		}
		if (player==5)
		{
			switch(Random_num(2))
			{
				case 0:
					pass(sock,off_posi[7],off_posi[9]);
					break;
				case 1:
					pass(sock,off_posi[9],off_posi[7]);
					break;
			}
		}
		if (player==6)
		{
			switch(Random_num(2))
			{
				case 0:
					pass(sock,off_posi[8],off_posi[10]);
					break;
				case 1:
					pass(sock,off_posi[10],off_posi[8]);
					break;
			}
			go_abs(sock,off_posi[player][0],off_posi[player][1]);
		}
		if (player==7)
		{
			if(mx>=25.0)
				pass(sock,goal_posi[1],NULL);
			else if (by>-34.0 && by<34.0)
			{
				if(Get_dist(mx,my,bx+1,by+1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by+1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx+1,by)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx,by+1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx;
					blocation[1]=by+1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx+1,by-1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by-1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx,by-1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx;
					blocation[1]=by-1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else
				{
					switch(Random_num(2))
					{
						case 0:
							pass(sock,off_posi[9],off_posi[10]);
							break;
						case 1:
							pass(sock,off_posi[10],off_posi[9]);
							break;
					}
					go_abs(sock,off_posi[player][0],off_posi[player][1]);
				}
			}
			else
			{
				switch(Random_num(2))
				{
					case 0:
						pass(sock,off_posi[9],off_posi[10]);
						break;
					case 1:
						pass(sock,off_posi[10],off_posi[9]);
						break;
				}
				go_abs(sock,off_posi[player][0],off_posi[player][1]);
			}
		}
		if (player==8)
		{
			if (mx>=25.0)
				pass(sock,goal_posi[1],NULL);
			else if (by>-34.0 && by<34.0)
			{
				if(Get_dist(mx,my,bx+1,by-1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by-1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx+1,by)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx,by-1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx;
					blocation[1]=by-1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx+1,by+1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by+1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx,by+1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx;
					blocation[1]=by+1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else
				{
					switch(Random_num(2))
					{
						case 0:
							pass(sock,off_posi[9],off_posi[10]);
							break;
						case 1:
							pass(sock,off_posi[10],off_posi[9]);
							break;
					}
					go_abs(sock,off_posi[player][0],off_posi[player][1]);
				}
			}
			else
			{
				switch(Random_num(2))
				{
					case 0:
						pass(sock,off_posi[9],off_posi[10]);
						break;
					case 1:
						pass(sock,off_posi[10],off_posi[9]);
						break;
				}
				go_abs(sock,off_posi[player][0],off_posi[player][1]);
			}
		}
		if (player==9)
		{
			if(mx>=25.0)
				pass(sock,goal_posi[1],NULL);
			else if (by>-34.0 && by<34.0)
			{
				if(Get_dist(mx,my,bx+1,by+1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by+1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx+1,by)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx,by+1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx;
					blocation[1]=by+1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx+1,by-1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by-1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx,by-1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx;
					blocation[1]=by-1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else
				{
					pass(sock,off_posi[10],NULL);
					go_abs(sock,off_posi[player][0],off_posi[player][1]);
				}
			}
			else
			{
				pass(sock,off_posi[10],NULL);
				go_abs(sock,off_posi[player][0],off_posi[player][1]);
			}
		}
		if (player==10)
		{
			if (mx>=25.0)
				pass(sock,goal_posi[1],NULL);
			else if (by>-34.0 && by<34.0)
			{
				if(Get_dist(mx,my,bx+1,by-1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by-1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx+1,by)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx,by-1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx;
					blocation[1]=by-1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx+1,by+1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx+1;
					blocation[1]=by+1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else if(Get_dist(mx,my,bx,by+1)!=oplayer[op_num].d[0])
				{
					blocation[0]=bx;
					blocation[1]=by+1;
					pass(sock,blocation,NULL);
					go_abs(sock,blocation[0],blocation[1]);
				}
				else
				{
					pass(sock,off_posi[9],NULL);
					go_abs(sock,off_posi[player][0],off_posi[player][1]);
				}
			}
			else
			{
				pass(sock,off_posi[9],NULL);
				go_abs(sock,off_posi[player][0],off_posi[player][1]);
			}
		}
//		}
//		else if ((my>(PENALTY_AREA_WIDTH+5)) || (my<(-PENALTY_AREA_WIDTH-5))) /* right & left flanks */
//		{
//	    		if (my>(PENALTY_AREA_WIDTH+5)) /* in the bottom area */
//				i=0;
//	    		else /* in the top area */
//				i=1;
//
//	    		if (mx<-25.0)
//				pass(sock,off_posi[8-i],plocation); /* pass to desired winger */
//			else if (mx>45) /* cross */
//          	 	{
//				if (player!=(10-i)) /* player passing != player recv. pass */
//					pass(sock,off_posi[10-i],plocation); /* pass to right fw */
//				else
//					pass(sock,off_posi[9+i],plocation); /* pass to left fw */
//           		}
//	    		else if (mx>20.0) 
//             			pass(sock,goal_posi[1],NULL); /* pass to opponent's goal */
//	    		else if (mx>0.0)
//				pass(sock,off_posi[10-i],off_posi[9+i]); /* pass to fw */
//	    		else
//			{
//				if (Random_num(2)==0) /* 0-winger, 1-fw */
//				{
//					if (Random_num(2)==0) /* 0-left, 1-right winger */
//               				     	pass(sock,off_posi[7+i],off_posi[8-i]); /* pass to left winger */
//                			else
//                    				pass(sock,off_posi[8-i],off_posi[7+i]); /* pass to right winger */
//				}
//				else
//				{
//					if (Random_num(2)==0)
//                       			 	pass(sock,off_posi[9+i],off_posi[10-i]); /* pass to left fw */
//                    			else
//                       				pass(sock,off_posi[10-i],off_posi[9+i]); /* pass to right fw */
//				}
//			}
//		} 
//		else
//		{
//			if (mx<-20.0) 
//           		{
//				/* generating random numbers for passing */
//				rand_num1=Random_num(5);
//				rand_num2=Random_num(4);
//				if (rand_num2>=rand_num1)
//					rand_num2++;
//
//                		pass(sock,off_posi[rand_num1+4],off_posi[rand_num2+4]); /* pass to mf */
//            		}
//			else if (mx<8.0)
//            		{
//				/* generating random numbers for passing */
//               	 		rand_num1=Random_num(4);
//                		rand_num2=Random_num(3);
//                		if (rand_num2>=rand_num1)
//                    			rand_num2++;
//
//				pass(sock,off_posi[rand_num1+7],off_posi[rand_num2+7]); /* pass to wingers or fw */
//            		}
//	    		else if (mx<18.0) 
//			{
//				/* generating random numbers for passing */
//				rand_num1=Random_num(2);
//                		rand_num2=0;
//                		if (rand_num2>=rand_num1)
//                    			rand_num2++;
//
//				pass(sock,off_posi[rand_num1+9],off_posi[rand_num2+9]); /* pass to fw */
//			}
//			else
//				pass(sock,goal_posi[1],NULL); /* tries a shot */
//		}
    	}
    	return TRUE;
}	

/* ************************************************************************** *
 * see() - visual sensor info. from server to client.
 *
 * Args:
 *		buf    : message buffer - (see Time ObjInfo ObjInfo ...).
 * 		t_name : team name.
 * Return type:
 *		void.
 * ************************************************************************** */
void see(char *buf, char *t_name)
{
	int  				i; /* 0-left, 1-right, 2-top, 3-bottom */
	char 				*b_ptr=buf,name[100],*n_ptr;
	char 				flag_info[15];
	int  				flag_counter;
	double 				odir;
	/* dummy===>for player whose teamname & unum cannot be seen */
	struct player_relative_address 	dummy,*dum_p;
	char 				*flag_strings[]=FLAG_STRINGS;
	FlagIndex 			findex;

	mp_num=0;
	op_num=0;

	if (++time_counter==MAX_TIME_COUNTER)
		time_counter=0;
	do
		++b_ptr;
	while((*b_ptr!='(') && (*b_ptr!=')'));
   	if (*b_ptr==')')
		return;
	b_ptr+=2;
	while(1)
	{
		switch(*b_ptr)
		{
			case 'l': /* line */
				b_ptr+=5;
				switch(*b_ptr)
				{
					case 'l': /* left end */
						i=0; break;
					case 'r': /* right end */
						i=1; break;
					case 't': /* top end */
						i=2; break;
					case 'b': /* bottom end */
						i=3; break;
				}
				if (side!='l')
					i^=0x1; /* Exclusive OR */
				b_ptr+=3;
				b_ptr=get_para(&line[i],b_ptr); /* points to the next ObjInfo */

				if (line[i].d[1]>0.0)
					line[i].d[1]=90.0-line[i].d[1];
				else
					line[i].d[1]=-90.0-line[i].d[1];

		    		break;

			case 'p': /* player */
				b_ptr+=6;
				if (*b_ptr==' ')
				{
					b_ptr++;

					n_ptr=name;
					while((*b_ptr!=')') && (*b_ptr!=' '))
						*n_ptr++=*b_ptr++;
					*n_ptr='\0';

					if (*b_ptr==' ')	/* Ignore jersey no. */
						while(*b_ptr!=')')
							b_ptr++;

					if (strcmp(name,t_name)==0)
						dum_p=&mplayer[mp_num++];
					else
						dum_p=&oplayer[op_num++];
					b_ptr=get_player_para(dum_p,b_ptr+2);

					if (dum_p->d[0]<2.0) /* collision check */
					{
						if ((dum_p->d[1]<60.0) && (dum_p->d[1]>0.0))
							collision_flag=-1;
						else if ((dum_p->d[1]>-60.0) && (dum_p->d[1]<0.0))
							collision_flag=1;
					}
				}
				else
				{
					b_ptr=get_player_para(&dummy,b_ptr+2);
				}
				break;

			case 'g': /* goal */
				b_ptr+=5;
	    			if (*b_ptr==side)
					i=0;
				else
					i=1;
				b_ptr=get_para(&goal[i],b_ptr+3);
				break;

	  		case 'b': case 'B': /* ball */
	    			b_ptr=get_para(&ball,b_ptr+6);
				b_flag=1;
	    			break;

			case 'f': /* flag */
				flag_counter=0;
				while (*b_ptr!=')')
				{
					flag_info[flag_counter++]=*b_ptr;
					b_ptr++;
				}
				flag_info[flag_counter]='\0';
				
				for (findex=0; flag_strings[findex]!=NULL; findex++)
					if (!strcmp(flag_strings[findex],flag_info))
					{
						if (side=='l')
							i=findex;
						else
							if ((findex<=26) && (findex!=0))
								i=findex+26;
							else
								i=findex-26;
						break;
					}

				b_ptr=get_para(&flag[i],b_ptr+2);
				break;
			case 'G': case 'F': case 'P': /* skip ObjInfo sensed by short
											 distance sensor */
				i=0;
				while (i<2)
				{
					b_ptr++;
					if (*b_ptr==')')
						i++;
				}
				b_ptr++;
				break;
		}
		if (*b_ptr==')') /* End of SEE info */
			break;
		else
			b_ptr+=3; /* move to the next ObjInfo */
	}
}
	
/* ************************************************************************** *
 * hear() - audible sensor info.
 *
 * Args:
 *		buf    : message buffer.
 *		t_name : team name.
 * Return type:
 * 		Bool : 	TRUE  - hear "kick_off_" or "play_on" message from referee.
 *		   	FALSE - other than "kick_off_" or "play_on" message.
 * ************************************************************************** */
Bool hear(char *buf, char *t_name)
{
    	int  		t_code,t;
    	char 		dir_str[50],mes[50];
    	double 		x,y;
	int 		i; /* for searching purpose */
	static char 	kick_off_side;
	static 		offside_cnt=0;
    
	sscanf(buf,"(hear %d %s %s)",&t,dir_str,mes);
	
	/* Search & replace ')' with '\0' */
	for (i=0; mes[i]!=')'; i++);
	mes[i] = '\0';

    	if (!strcmp(mes,t_name)) /* message from team mate or SELF. */
	{
		sscanf(buf,"(hear %d %s %s %d %lf %lf)",&t,dir_str,mes,&t_code,&bx,&by);
		if (t_code>=time_stamp)
		{
		    	time_stamp=t_code;
		    	if ((!b_flag) && (strcmp(dir_str,SELF)!=0))
				h_flag=1;
		}
		else if (strcmp(dir_str,PLAYER))
				p_flag=1; 
/*		else
		    printf("mimic message \n  %s!!\n",buf);*/
    	} 
	else if (!strcmp(dir_str,REFEREE))
	{
		/* half time & time extended ===> move to position */
		if ((!strcmp(mes,"half_time")) || (!strcmp(mes,"time_extended"))
			|| (!strncmp(mes,"goal_l",6)) || (!strncmp(mes,"goal_r",6)))
			move_flag=1;
		if ((!strncmp(mes,"offside_",8)) && (mes[8]==side))
       	 	{
			if (((player==9) || (player==10)) && (bx<(off_posi[player][0]-10.0)))
			{
				if (Abs_d(mx-off_posi[player][0])<3)
					offside_cnt++;
				if ((offside_cnt==1) || (offside_cnt%2==0))
				{
            				off_posi[9][0] -=5.0;
					off_posi[10][0]-=5.0;
				}
			}
        	}
		if ((!strncmp(mes,"corner_kick_",12)) && (mes[12]==side))
        	{
            		off_posi[9][0] =30.0;
            		off_posi[10][0]=30.0;
       	 	}
		if (!strncmp(mes,"goalie_catch_ball_",18))
		{
			if (mes[18]==side)
				catch_flag=TRUE;
		}
		if (!strncmp(mes,"kick_off_",9))
		{
			kick_off_side=mes[9];
			if (side==kick_off_side)
				return TRUE;
		}
		if ((!strcmp(mes,"play_on")) && (kick_off_side!=side))
			return TRUE;
    	}
    	return FALSE;
}
	
/* ************************************************************************** *
 * message_loop().
 *
 * Args:
 * 		sock   : Socket.
 * 		t_name : team name.
 *		fpin   : input file.
 *		fpout  : output file.
 * Return type:
 *		void.
 * ************************************************************************** */
void message_loop(Socket sock, char *t_name, FILE* fpin, FILE* fpout)
{
    	fd_set	readfds ;
    	int	in, max_fd, n ;
    	char	buf[MAXMESG];
    	char	end_mes[]="quit";
	char 	sensor_info[16];
	int 	think_value;	/* value return by think() */

    	in = fileno(fpin);

    	while(1)
	{
		FD_ZERO(&readfds) ;
		FD_SET(in, &readfds) ;
		FD_SET(sock.socketfd, &readfds) ;

		max_fd = ((in > sock.socketfd) ? in : sock.socketfd) + 1 ;
		if(select(max_fd, &readfds, 0, 0, 0) == -1)
		{
		    	perror("select") ;
		    	break ;
		}

		if(FD_ISSET(in, &readfds))
		{
			printf("fgets: ");
		    	fgets(buf, MAXMESG, fpin) ;
		    	if (strncmp(end_mes,buf,4)==0)
			break;
		}

		if (FD_ISSET(sock.socketfd, &readfds))	
		{
	    		n = receive_message(&sock, buf, MAXMESG);
	    		if(n == -1)
				break ;
	    		else if(n > 0)
			{
				sscanf(buf, "(%s ", sensor_info);
				
	    			if (!strcmp(sensor_info, SEE))
				{
		    			see(buf,t_name); /* SEE sensor command */	
		    			s_flag=1;		    
	    			} 
				else if (!strcmp(sensor_info, HEAR))
		    			hear(buf,t_name); /* HEAR sensor command */
				else if (!strcmp(sensor_info, SENSE_BODY))
                    			printf("player%d===>%s\n",player,buf); /* sense_body sensor command */
				else /* error */
					break;
				fflush(stdout) ;
	    		}			
		}

		switch (player)
		{
			case 0:
				think_value=gk_think(&sock,t_name);
				break;
			default:
				think_value=think(&sock,t_name);
				break;
		}

		if (think_value)
		{
	    		s_flag=0;
	    		h_flag=0;
	    		b_flag=0;
			collision_flag=0;
		}
    	}
}

/* ************************************************************************** * 
 * main().
 *
 * Args:
 * 		argc : No. of arguments.
 * 		argv : arguments [optional].
 *	       	argv[0] - this executable filename, ie. KRDL.
 *	       	argv[1] - team name.
 *	        argv[2] - server.
 *	        argv[3] - port number.
 * Return type:
 *		void.
 * ************************************************************************** */
#ifdef TEST	/* for testing */
void main(int argc, char **argv)
{
    	Socket 	sock;
	int 	goalie_flag;
    	char 	team_name[16];
    	char 	*server;
    	int 	port;
	char 	buf[MAXMESG];

	if (argc == 5)
    	{
		if (!strcmp(argv[1],"goalie"))
			goalie_flag=TRUE;
		else
			goalie_flag=FALSE;
       	 	strcpy(team_name,argv[2]);
        	server = argv[3];
        	port = atoi(argv[4]);
    	}
    	else if (argc == 4)
    	{
		if (!strcmp(argv[1],"goalie"))
		{
       	     		goalie_flag=TRUE;
			strcpy(team_name,argv[2]);
        		server = argv[3];
        		port = 6000;
		}
        	else
		{
            		goalie_flag=FALSE;
    			strcpy(team_name,argv[1]);
			server = argv[2];
			port = atoi(argv[3]);
		}
    	}
    	else if(argc == 3) 
    	{
		if (!strcmp(argv[1],"goalie"))
		{
            		goalie_flag=TRUE;
			strcpy(team_name,argv[2]);
        		server = "localhost";
        		port = 6000;
		}
        	else
		{
            		goalie_flag=FALSE;
			strcpy(team_name,argv[1]);
			server = argv[2];
			port = 6000;
		}
    	} 
    	else if (argc == 2) 
    	{
		if ((!strcmp(argv[1], "/?")) || (*argv[1]=='?')) /* Help */
		{
			printf("\nUsage:\n");
			printf("\t%s [Team_name [Server [Port]]]\n", argv[0]);
		}
		else
		{
			if (!strcmp(argv[1],"goalie"))
			{
       	     			goalie_flag=TRUE;
				strcpy(team_name,"KRDL");
            			server = "localhost";
            			port = 6000;
			}
        		else
			{
            			goalie_flag=FALSE;
				strcpy(team_name,argv[1]);
				server = "localhost";
				port = 6000;
			}
		}
    	} 
    	else	/* default settings */ 
    	{
		goalie_flag=FALSE;
		strcpy(team_name,"KRDL");	
		server = "localhost";
		port = 6000;
    	}
    
    	sock = init_connection(server,port);
    
    	if (sock.socketfd == -1)
		exit(1);
    
    	if (init_condition(&sock, team_name, goalie_flag))
	    	message_loop(sock, team_name, stdin, stdout);

    	close_connection(sock);
}
#endif /* TEST */

